/***************************************************************************
 *
 * xxamp.c - C source for xAMp protocol suite
 *
 * Author: Werner Vogels & Antonio Casimiro Costa
 *
 * Copyright (c) 1997 Navigators Group
 *
 * This file is part of Navigators Group xAMp protocol.
 *
 * This software is copyrighted by the Navigators Group of Faculty of
 * Sciences of University of Lisbon and SHOULD ONLY BE USED in
 * the context of the Licence Agreement presented in the file LICENCE.
 *
 * xAMp is distributed WITHOUT ANY WARRANTY. See the Navigators Group
 * Software License for more details.
 *
 * You should have received a copy of the Navigators Group Software License
 * along with xAMp; see the file LICENCE.  If not, write to:
 *       Departamento de Informatica, FCUL,
 *       Bloco C5 Campo Grande, 1700 LISBOA -- PORTUGAL
 * or fax: +351 1 7500084
 * or mail: navigators@di.fc.ul.pt
 *
 * For more information see the file README or send e-mail to:
 * 	navigators@di.fc.ul.pt
 *
 ***************************************************************************/


/*------------------------------------------------------------------------+
|   Includes
+-------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <netinet/in.h>        /* Byte order convertions */

#include "user/xampuser.h"

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>     /* Get standard string definations. */

#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/TextSrc.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Sme.h>
#include <X11/Xaw/SmeBSB.h>

#include <X11/bitmaps/gray1>

#include "mgsstate_empty"
#include "mgsstate_active"
#include "mgsstate_inactive"

/*------------------------------------------------------------------------+
|   External functions
+-------------------------------------------------------------------------*/

extern int     exit ();

/*------------------------------------------------------------------------+
|   Definitions
+-------------------------------------------------------------------------*/

#define MAX_STATIONS         9      /* We only show up to 10 stations */
#define MAX_ACTIONS         20      /* Maximum 20 simultaneous actions */

#define GROUP_OPEN           1      /* Actions */
#define GROUP_CLOSE          2

/*------------------------------------------------------------------------+
|   Global variables
+-------------------------------------------------------------------------*/

String fallback_resources[] = {
    "*font: -*-helvetica-medium-r-*-*-12-*-*-*-*-*-*-*",
    "*mbutton.font: -*-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*",
    "*timelabel.font: -*-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*",
    "*entry.font: -*-lucida-bold-r-*-*-12-*-*-*-*-*-*-*",
    "*paned.height: 193",
    "*infotext*font: fixed",
    "*datatext*font: fixed",
    "*datatext*width: 50",
    "*datatext*height: 19",
    "*datalabel.width: 40",
    "*datalabel.borderWidth: 0",
    "*datalabel.justify: XtJustifyRight",
    "*button.width: 55",
    "*output.width: 60",
    "*mbutton.width: 50",
    "*infotext.width: 538",
    "*infotext.height: 100",
    "*timelabel*width: 45",
    "*xxamp.defaultDistance: 4",
    "*form.defaultDistance: 4",
    "*menu*width: 70",
    "*entry.vertSpace: 40",
    NULL,
};

static XrmOptionDescRec options[] = {
    {"-vspace", "*Box.vSpace", XrmoptionSepArg, NULL},
};

Pixmap gray1, emptypix, activepix, inactivepix;

Widget mgsstate[MAX_STATIONS];
Widget mgstext;

Widget indresultbutton, confresultbutton, conflabel, indlabel;

Widget timenonebutton, timeappbutton, timedaemonbutton;

Widget groupstateview[MAXGRP], groupstatelabel[MAXGRP];

Widget groupidbutton, objidbutton, sizebutton, priobutton;
Widget tstimelabel;

typedef struct GroupEntrie {
    int    group;
    ObjUid obj[MAXOBJ];           
    int    nobj;
} GroupEntrie;

GroupEntrie grouptable[MAXGRP];   

typedef struct ObjAction {    /* Used for group open and close */
    ObjUid obj;
    int    action;
} ObjAction;

ObjAction  action[MAX_ACTIONS]; 

ObjUid last_oid;  
ObjUid mgsobj;     

int mgs_active = 0;

static int timing = T_NONE;   /* For timing debugging */


/*----------------------------------------------------------------------
|
|   General routines
|
+----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
|   Function: printId
+----------------------------------------------------------------------*/

char* printId (id)
{
    switch (id)
    {
    case RELIABLE_SEND:
	return "RELIABLE_SEND";
    case RELIABLE_RECV:
	return "RELIABLE_RECV";
    case ATOMIC_SEND:
	return "ATOMIC_SEND";
    case ATOMIC_RECV:
	return "ATOMIC_RECV";
    case TIGHT_SEND:
	return "TIGHT_SEND";
    case TIGHT_RECV:
	return "TIGHT_RECV";
    case DELTA_SEND:
	return "DELTA_SEND";
    case DELTA_RECV:
	return "DELTA_RECV";
    case ATLEASTN_SEND:
	return "ATLEASTN_SEND";
    case ATLEASTTO_SEND:
	return "ATLEASTTO_SEND";
    case ATLEAST_RECV:
	return "ATLEAST_RECV";
    case BESTEFFN_SEND:
	return "BESTEFFN_SEND";
    case BESTEFFTO_SEND:
	return "BESTEFFTO_SEND";
    case BESTEFF_RECV:
	return "BESTEFF_RECV";
    case GATE_REQUEST:
	return "GATE_REQUEST";
    case NEW_VIEW:
	return "NEW_VIEW";
    case OPEN_REQUEST:
	return "OPEN_REQUEST";
    case CLOSE_REQUEST:
	return "CLOSE_REQUEST";
    case MGS_JOIN:
	return "MGS_JOIN";
    case MGS_LEAVE:
	return "MGS_LEAVE";
    case MGS_CHECK:
	return "MGS_CHECK";
    case MGS_NEW:
	return "MGS_NEW";
    default:
	return "UNKOWN";
    }
}

/*----------------------------------------------------------------------
|   Function: putText
+----------------------------------------------------------------------*/

void putText (w, str)
    Widget w;
    char*  str;
{
    XawTextPosition length;
    XawTextBlock    txt;

    txt.length   = strlen(str);
    txt.ptr      = str;
    txt.firstPos = 0;
    txt.format   = FMT8BIT;
    
    length = XawTextGetInsertionPoint (w);
    
    XawTextReplace(w, (XawTextPosition)0, length, &txt);
    XawTextSetInsertionPoint (w, txt.length);
}

/*----------------------------------------------------------------------
|   Function: printMess
+----------------------------------------------------------------------*/

void printMess (str)
    char* str;
{
    XawTextPosition length;
    XawTextBlock    txt;

    txt.length   = strlen(str);
    txt.ptr      = str;
    txt.firstPos = 0;
    txt.format   = FMT8BIT;
    
    length = XawTextGetInsertionPoint (mgstext);
    
    XawTextReplace(mgstext, length, length, &txt);
}

/*----------------------------------------------------------------------
|   Function: printMessInfo
+----------------------------------------------------------------------*/

void printMessInfo (type, id, stat, group, objname)
    char*   type;
    short   id;
    short   stat;
    int     group;
    char*   objname;
{
    char            str[4096];

    if (objname)
	sprintf (str,"type %s id %s group %d oid %s status %s\n",
		 type, printId(id), group, objname, stat ? "NOT OK":"OK");
    else
	sprintf (str,"type %s id %s group %d status %s\n",
		 type, printId(id), group, stat ? "NOT OK":"OK");
    printMess (str);
}

/*----------------------------------------------------------------------
|   Function: initActionEntries
+----------------------------------------------------------------------*/

void initActionEntries ()
{
    int i;

    for (i=1; i<MAX_ACTIONS; i++)
	action[i].obj = ouidNew ();
}

/*----------------------------------------------------------------------
|   Function: getActionEntry
+----------------------------------------------------------------------*/

int getActionEntry ()
{
    int i;

    for (i=1; i<MAX_ACTIONS; i++)
	if (ouidIsNull (action[i].obj))
	    return i;
    return -1;
}

/*----------------------------------------------------------------------
|   Function: initGroupEntries
+----------------------------------------------------------------------*/

void initGroupEntries ()
{
    int i, j;

    for (i=1; i<MAXGRP; i++)
    {
	grouptable[i].group = -1;
	for (j=1; j<MAXOBJ; j++)
	    grouptable[i].obj[j] = ouidNew ();
    }
}

/*----------------------------------------------------------------------
|   Function: findGroup
+----------------------------------------------------------------------*/

int findGroup (group)
    int group;
{
    int i;

    for (i=1; i<MAXGRP; i++)
	if (grouptable[i].group == group) 
	    return i;
    return -1;
}
   
/*----------------------------------------------------------------------
|   Function: getGroupEntry
+----------------------------------------------------------------------*/

int getGroupEntry ()
{
    int i;

    for (i=1; i<MAXGRP; i++)
	if (grouptable[i].group == -1) 
	    return i;
    return -1;
}

/*----------------------------------------------------------------------
|   Function: findObj
+----------------------------------------------------------------------*/

int findObj (groupE, obj)
    int    groupE;
    ObjUid obj;
{
    int i;

    for (i=1; i<MAXOBJ; i++)
	if (ouidIsEqual (grouptable[groupE].obj[i], obj)) 
	    return i;
    return -1;
}
   
/*----------------------------------------------------------------------
|   Function: getObjEntry
+----------------------------------------------------------------------*/

int getObjEntry (groupE)
    int    groupE;
{
    int i;

    for (i=1; i<MAXOBJ; i++)
	if (ouidIsNull (grouptable[groupE].obj[i])) 
	    return i;
    return -1;
}
   
/*----------------------------------------------------------------------
|   
|   Handler routines
|
+----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
|   Function: mgsHandler
+----------------------------------------------------------------------*/

void mgsHandler (short id, short stat, SiteEntry* st)
{
    Arg  args[10];
    char name[30];
    int  i, j;

    if (id == MGS_NEW)
	printMessInfo ("INDI", id, stat, 0, NULL);
    else
	printMessInfo ("CONF", id, stat, 0, NULL);
    
    if (stat != STATUS_OK)
	return;
	
    switch (id) {
    case MGS_JOIN: mgs_active = 1;
    case MGS_NEW:
    case MGS_CHECK:
	for (i=0, j=31; i<MAX_STATIONS; i++,j--) 
	{
	    if (st[j].used)
		if (st[j].active)
		    XtSetArg(args[0],XtNleftBitmap,activepix);
		else
		    XtSetArg(args[0],XtNleftBitmap,inactivepix);
	    else
		XtSetArg(args[0],XtNleftBitmap,emptypix);
	    if (st[j].used) 
	    {
#ifdef sun
		if (ether_ntohost(name,&st[j].pname))
		    strcpy(name,ether_ntoa(&st[j].pname));
#else
		st[j].pname.hh = ntohs (st[j].pname.hh);
		st[j].pname.hl = ntohs (st[j].pname.hl);
		st[j].pname.ll = ntohs (st[j].pname.ll);
	        sprintf (name,"%2x:%2x:%2x:%2x:%2x:%2x",
			 ((char*)(&st[j].pname))[0] & 0xff,
			 ((char*)(&st[j].pname))[1] & 0xff,
			 ((char*)(&st[j].pname))[2] & 0xff,
			 ((char*)(&st[j].pname))[3] & 0xff,
			 ((char*)(&st[j].pname))[4] & 0xff,
			 ((char*)(&st[j].pname))[5] & 0xff);
#endif
	    }
	    else
		strcpy(name,"-");
	    XtSetArg(args[1],XtNlabel,name);
	    XtSetValues(mgsstate[i],args,TWO);
	}
	break;
	
    case MGS_LEAVE:
	for (i=0; i<MAX_STATIONS; i++) 
	{
	    XtSetArg(args[0],XtNleftBitmap,emptypix);
	    XtSetArg(args[1],XtNlabel,"-");
	    XtSetValues(mgsstate[i],args,TWO);
	}
	xampTiming (mgsobj, 0);
	mgs_active = 0;
	break;

    default:
	break;
    }
}

/*----------------------------------------------------------------------
|   Function: viewHandler
+----------------------------------------------------------------------*/

void viewHandler (int group, OuidList* view)
{
    char vw[1000];
    char str[100];
    Arg  args[5];
    int  i, j;
    
    printMessInfo ("INDI", NEW_VIEW, STATUS_OK, group, NULL);

    i = findGroup (group);
    if ((i != -1) && (i < 7))
	{
	    sprintf (vw, "%d", group);
	    XtSetArg (args[0], XtNlabel, vw);
	    XtSetValues (groupstatelabel[i], args, ONE);
	    
	    if (ouidToString (ouidListElem (view, 0), str) == -1)
	    {
		printMess ("View handler received bad ouidList\n");
		return;
	    }
	    sprintf (vw, "%s", str);
	    for (j=1; j<ouidListSize(view); j++)
	    {
		if (ouidToString (ouidListElem (view, j), str) == -1)
		{
		    printMess ("View handler received bad ouidList\n");
		    return;
		}
		strcat (vw, ", ");
		strcat (vw, str);
	    }
	    XtSetArg (args[0], XtNlabel, vw);
	    XtSetValues (groupstateview[i], args, ONE);
	}
}

/*----------------------------------------------------------------------
|   Function: confHandler
+----------------------------------------------------------------------*/

void confHandler (int group, ObjUid obj, short id, short stat, char* data, 
		  int size, int key)
{
    char  vw[1000];
    char  str[100];
    Arg   args[5];
    int   i, j;

    ouidToString (obj, str);
    printMessInfo ("CONF", id, stat, group, str);

    if (stat == STATUS_NOT_OK)
    {
	if (id == OPEN_REQUEST)
	    printMess ("May be duplicated object id...\n");
	action[key].obj = ouidNew ();
	return;
    }

    if ((id == OPEN_REQUEST) || (id == CLOSE_REQUEST))
    {
	if (action[key].action == GROUP_CLOSE)
	{
	    i = findGroup (group);
	    if (i == -1)
	    {
		printMess ("Error: group not found.\n");
		action[key].obj = ouidNew ();
		return;
	    }
	    j = findObj (i, action[key].obj);
	    if (j == -1)
	    {
		printMess ("Error: obj not found.\n");
		action[key].obj = ouidNew ();
		return;
	    }
	    action[key].obj = ouidNew ();
	    grouptable[i].obj[j] = ouidNew ();
	    grouptable[i].nobj --;
	    /*
	      Should do timing(off) if this is last object
	    */
	    if (grouptable[i].nobj == 0)
	    {
		grouptable[i].group = -1;
		if (i < 7)
		{
		    XtSetArg (args[0], XtNlabel, "");
		    XtSetValues (groupstatelabel[i], args, ONE);
		    XtSetValues (groupstateview[i], args, ONE);
		}
	    }
	}
	else
	{
	    if ((i = findGroup (group)) == -1)
	    {
		if ((i = getGroupEntry() ) != -1)
		    grouptable[i].group = group;
		else
		{
		    printMess ("Error: group table full.\n");
		    action[key].obj = ouidNew ();
		    return;
		}
	    }
	    j = getObjEntry (i);
	    if (j == -1)
	    {
		printMess ("Error: obj table full.\n");
		action[key].obj = ouidNew ();
		return;
	    }
	    grouptable[i].obj[j] = action[key].obj;
	    grouptable[i].nobj ++;
	    action[key].obj = ouidNew ();
	}
    }
}

/*----------------------------------------------------------------------
|   Function: dataHandler
+----------------------------------------------------------------------*/

void dataHandler (int group, ObjUid obj, short id, char* info, char* data, 
		  int size)
{
    char str[100];

    ouidToString (obj, str);
    printMessInfo ("DATA", id, STATUS_OK, group, str);
}

/*----------------------------------------------------------------------
|   Function: timingHandler
+----------------------------------------------------------------------*/

void timingHandler (short type, TimeVal timestamp)
{
    static TimeVal  starttime;
    static int      ctimed = 0;
    static int      itimed = 0;
    long sec, msec;
    Arg  args[10];
    char str[4096];

    if (type == IO_TIME)  
    {
	ctimed = 1;
	itimed = 1;
	starttime = timestamp;
	sprintf (str, "0.0");
	XtSetArg (args[0], XtNlabel, str);
	XtSetValues (confresultbutton, args, ONE);
	XtSetValues (indresultbutton, args, ONE);
	return;
    } 

    if (timing && (ctimed || itimed)) 
    {
	sec  = timestamp.tv_sec - starttime.tv_sec;
	msec = timestamp.tv_usec - starttime.tv_usec;
	
	if (msec < 0)
	{
	    sec--;
	    msec += 1000000;
	}
	
	sprintf (str, "%d.%06d", sec, msec);
	XtSetArg (args[0], XtNlabel, str);
	switch (type) {
	case IO_DATA:
	    XtSetValues (indresultbutton, args, ONE);
	    itimed = 0;
	    break;
	case IO_CONFIRMATION:
	    XtSetValues (confresultbutton, args, ONE);
	    ctimed = 0;
	    break;
	case IO_INDICATION:
	    XtSetValues (indresultbutton, args, ONE);
	    itimed = 0;
	    break;
	default:
	    printMess ("Unknown head type\n");
	    break;
	}
    }
}

/*----------------------------------------------------------------------
|   Callback routines
+----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
|   Function: timingNoneCallb
+----------------------------------------------------------------------*/

void timingNoneCallb ()
{
    Arg args[1];
    int i, j;

    for (i=1; i<MAXGRP; i++)
    {
	if (grouptable[i].group != -1)
	    for (j=1; j<MAXOBJ; j++)
		if (!ouidIsNull (grouptable[i].obj[j]))
		    xampTiming (grouptable[i].obj[j], 0);
    }
    timing = T_NONE;
    XtSetArg (args[0], XtNsensitive, False);
    XtSetValues (timenonebutton, args, ONE);

    XtSetArg (args[0], XtNsensitive, False);
    XtSetValues (indlabel, args, ONE);

    XtSetArg (args[0], XtNsensitive, False);
    XtSetArg (args[1], XtNlabel, "0.0");
    XtSetValues (indresultbutton, args, TWO);

    XtSetArg (args[0], XtNsensitive, False);
    XtSetValues (conflabel, args, ONE);
 
    XtSetArg (args[0], XtNsensitive, False);
    XtSetArg (args[1], XtNlabel, "0.0");
    XtSetValues (confresultbutton, args, TWO);

    XtSetArg (args[0], XtNsensitive, True);
    XtSetValues (timeappbutton, args, ONE);
    XtSetArg (args[0], XtNsensitive, True);
    XtSetValues (timedaemonbutton, args, ONE);
}

/*----------------------------------------------------------------------
|   Function: timingAppCallb
+----------------------------------------------------------------------*/

void timingAppCallb ()
{
    Arg args[1];
    int i, j;

    if (timing == T_NONE) 
    {
	for (i=1; i<MAXGRP; i++)
	{
	    if (grouptable[i].group != -1)
		for (j=1; j<MAXOBJ; j++)
		    if (!ouidIsNull (grouptable[i].obj[j]))
		    {
			xampTiming (grouptable[i].obj[j], timingHandler);
			xampTimingMode (grouptable[i].obj[j], T_APP);
		    }
	}
	if (mgs_active)
	{
	    xampTiming (mgsobj, timingHandler);
	    xampTimingMode (mgsobj, T_APP);
	}
	XtSetArg (args[0], XtNsensitive, True);
	XtSetValues (timenonebutton, args, ONE);
	XtSetValues (indlabel, args, ONE);
	XtSetValues (indresultbutton, args, ONE);
	XtSetValues (conflabel, args, ONE);
	XtSetValues (confresultbutton, args, ONE);
	XtSetArg (args[0], XtNsensitive, False);
	XtSetValues (timeappbutton, args, ONE);
    }
    else
    {
	for (i=1; i<MAXGRP; i++)
	{
	    if (grouptable[i].group != -1)
		for (j=1; j<MAXOBJ; j++)
		    if (!ouidIsNull (grouptable[i].obj[j]))
			xampTimingMode (grouptable[i].obj[j], T_APP);
	}
	XtSetArg (args[0], XtNsensitive, True);
	XtSetValues (timedaemonbutton, args, ONE);
	XtSetArg (args[0], XtNsensitive, False);
	XtSetValues (timeappbutton, args, ONE);
    }
    timing = T_APP;
}

/*----------------------------------------------------------------------
|   Function: timingDaemonCallb
+----------------------------------------------------------------------*/

void timingDaemonCallb ()
{
    Arg args[1];
    int i, j;

    if (timing == 0) 
    {
	for (i=1; i<MAXGRP; i++)
	{
	    if (grouptable[i].group != -1)
		for (j=1; j<MAXOBJ; j++)
		    if (!ouidIsNull (grouptable[i].obj[j]))
		    {
			xampTiming (grouptable[i].obj[j], timingHandler);
			xampTimingMode (grouptable[i].obj[j], T_XAMP);
		    }
	}
	if (mgs_active)
	{
	    xampTiming (mgsobj, timingHandler);
	    xampTimingMode (mgsobj, T_XAMP);
	}
	XtSetArg (args[0], XtNsensitive, True);
	XtSetValues (timenonebutton, args, ONE);
	XtSetValues (indlabel, args, ONE);
	XtSetValues (indresultbutton, args, ONE);
	XtSetValues (conflabel, args, ONE);
	XtSetValues (confresultbutton, args, ONE);
	XtSetArg (args[0], XtNsensitive, False);
	XtSetValues (timedaemonbutton, args, ONE);
    }
    else
    {
	for (i=1; i<MAXGRP; i++)
	{
	    if (grouptable[i].group != -1)
		for (j=1; j<MAXOBJ; j++)
		    if (!ouidIsNull (grouptable[i].obj[j]))
			xampTimingMode (grouptable[i].obj[j], T_XAMP);
	}
	XtSetArg (args[0], XtNsensitive, True);
	XtSetValues (timeappbutton, args, ONE);
	XtSetArg (args[0], XtNsensitive, False);
	XtSetValues (timedaemonbutton, args, ONE);
    }
    timing = T_XAMP;
}

/*----------------------------------------------------------------------
|   Function: groupOpenCallb
+----------------------------------------------------------------------*/

void groupOpenCallb ()
{ 
    int    group;
    ObjUid oid;
    int    key;
    Arg    args[1];
    String str;
    char   str1[100];

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (groupidbutton, args, ONE);
    group = atoi (str);
    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (objidbutton, args, ONE);
    oid = ouidFromString (str);
  
    if (ouidIsEqual (oid, ouidBuild (0, 0)))
    {
	oid = ouidGetUniq ();
	ouidToString (oid, str1);
	putText (objidbutton, str1);
    }
    last_oid = oid;

    if (timing)
    {
	xampTiming (oid, timingHandler);
	xampTimingMode (oid, timing);
    }
    
    key = getActionEntry ();
    if (key == -1)
	printMess ("Error: Action table full.\n");
    else
    {
	action[key].action = GROUP_OPEN;
	action[key].obj    = oid;
    
	xampGroupOpen (group, oid, viewHandler, confHandler, dataHandler, key);
    }
}	

/*----------------------------------------------------------------------
|   Function: groupCloseCallb
+----------------------------------------------------------------------*/

void groupCloseCallb ()
{
    int    group;
    ObjUid oid;
    int    key;
    Arg    args[1];
    String str;
    char   str1[100];

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (groupidbutton, args, ONE);
    group = atoi (str);

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (objidbutton, args, ONE);
    oid = ouidFromString (str);

    if (ouidIsEqual (oid, ouidBuild (0, 0)))
    {
	oid = last_oid;
	ouidToString (oid, str1);
	putText (objidbutton, str1);
    }

    key = getActionEntry ();
    if (key == -1)
	printMess ("Error: Action table full.\n");
    else
    {
	action[key].action = GROUP_CLOSE;
	action[key].obj    = oid;

	if (xampGroupClose (group, oid, key) == -2)
	    printMess ("Group is not open.\n");
    }
}	

/*----------------------------------------------------------------------
|   Function: reliableCallb
+----------------------------------------------------------------------*/

void reliableCallb ()
{
    int    group;
    int    size;
    ObjUid oid;
    char   bb[4096];
    Arg    args[1];
    String str;
    char   str1[100];

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (groupidbutton, args, ONE);
    group = atoi (str);

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (sizebutton, args, ONE);
    size  = atoi (str);

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (objidbutton, args, ONE);
    oid = ouidFromString (str);

    if (ouidIsEqual (oid, ouidBuild (0, 0)))
    {
	oid = last_oid;
	ouidToString (oid, str1);
	putText (objidbutton, str1);
    }
    
    if (reliableSend (group, oid, ALL_OBJ, bb, size, 1, 0) == -2)
	printMess ("Must open group before sending.\n");
}

/*----------------------------------------------------------------------
|   Function: atomicCallb
+----------------------------------------------------------------------*/

void atomicCallb ()
{
    int    group;
    int    size;
    ObjUid oid;
    char   bb[4096];
    Arg    args[1];
    String str;
    char   str1[100];

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (groupidbutton, args, ONE);
    group = atoi (str);

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (sizebutton, args, ONE);
    size  = atoi (str);

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (objidbutton, args, ONE);
    oid = ouidFromString (str);

    if (ouidIsEqual (oid, ouidBuild (0, 0)))
    {
	oid = last_oid;
	ouidToString (oid, str1);
	putText (objidbutton, str1);
    }

    if (atomicSend (group, oid, ALL_OBJ, bb, size, 1, 0) == -2)
	printMess ("Must open group before sending.\n");
}

/*----------------------------------------------------------------------
|   Function: tightCallb
+----------------------------------------------------------------------*/

void tightCallb ()
{
    int    group;
    int    size;
    ObjUid oid;
    int    prio;
    char   bb[4096];
    Arg    args[1];
    String str;
    char   str1[100];

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (groupidbutton, args, ONE);
    group = atoi (str);

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (sizebutton, args, ONE);
    size  = atoi (str);

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (priobutton, args, ONE);
    prio  = atoi (str);

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (objidbutton, args, ONE);
    oid = ouidFromString (str);

    if (ouidIsEqual (oid, ouidBuild (0, 0)))
    {
	oid = last_oid;
	ouidToString (oid, str1);
	putText (objidbutton, str1);
    }

    if (tightSend (group, oid, ALL_OBJ, bb, size, 1, prio, 0) == -2)
	printMess ("Must open group before sending.\n");
}

/*----------------------------------------------------------------------
|   Function: deltaCallb
+----------------------------------------------------------------------*/

void deltaCallb ()
{
    int    group;
    int    size;
    ObjUid oid;
    char   bb[4096];
    Arg    args[1];
    String str;
    char   str1[100];

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (groupidbutton, args, ONE);
    group = atoi (str);

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (sizebutton, args, ONE);
    size  = atoi (str);

    XtSetArg (args[0], XtNstring, &str);
    XtGetValues (objidbutton, args, ONE);
    oid = ouidFromString (str);

    if (ouidIsEqual (oid, ouidBuild (0, 0)))
    {
	oid = last_oid;
	ouidToString (oid, str1);
	putText (objidbutton, str1);
    }

    if (deltaSend (group, oid, ALL_OBJ, bb, size, 0) == -2)
	printMess ("Must open group before sending.\n");
}

/*----------------------------------------------------------------------
|   Function: mgsJoinCallb
+----------------------------------------------------------------------*/

void mgsJoinCallb ()
{
    if (timing)
    {
	xampTiming (mgsobj, timingHandler);
	xampTimingMode (mgsobj, timing);
    }
    if (xampMgsJoin (mgsobj, mgsHandler) == -3)
	printMess ("MGS already joined.\n");
}

/*----------------------------------------------------------------------
|   Function: mgsLeaveCallb
+----------------------------------------------------------------------*/

void mgsLeaveCallb ()
{
    if (xampMgsLeave (mgsobj) == -2)
	printMess ("MGS not joined.\n");
}

/*----------------------------------------------------------------------
|   Function: mgsCheckCallb
+----------------------------------------------------------------------*/

void mgsCheckCallb ()
{
    if (xampMgsCheck (mgsobj) == -2)
	printMess ("MGS not joined.\n");
}

/*----------------------------------------------------------------------
|   Function: tsReadCallb
+----------------------------------------------------------------------*/

void tsReadCallb ()
{
    char    str[4096];
    Arg     args[10];
    TimeVal vtime;

    if (tsRead (&vtime) != 0)
	printMess ("Could not read synchronized time\n");
    else
    {
	sprintf(str, "%d.%06d", vtime.tv_sec, vtime.tv_usec);
	XtSetArg (args[0], XtNlabel, str);
	XtSetValues (tstimelabel, args, ONE);
    }
}

/*----------------------------------------------------------------------
|   Function: quitCallb
+----------------------------------------------------------------------*/

void quitCallb ()
{
    xampClose ();
    exit (0);
}

/*----------------------------------------------------------------------
|   Function: main
+----------------------------------------------------------------------*/

main (int argc, char* argv[])
{
    XtAppContext app_con;
    Widget toplevel, box, quitbutton;
    Widget mgsstatepaned;
    Widget mgsmenu, mgsmenubutton;
    Widget mgsjoinbutton, mgscheckbutton, mgsleavebutton;
    Widget timemenu, timemenubutton;
    Widget tsgetbutton;
    Widget timeform;
    Widget datactl, comctl;
    Widget grouplabel, objlabel, sizelabel, priolabel;
    Widget groupopenbutton, groupclosebutton;
    Widget reliable, atomic, tight, delta;
    Widget groupview, gvpanedlabel, gvpanedview;
    Arg    args[10];
    char   name[1024], hostname[1024];
    int    xamp_fd;
    int    i;
    unsigned int depth;

    initActionEntries ();
    initGroupEntries  ();
    mgsobj = ouidBuild (0L, 9999L);
    xamp_fd = xampOpen (1, 0, 0);
    
    if (xamp_fd < 0)
    {
	printf ("Error opening xAMp.\n");
	exit (1);
    }
    gethostname (hostname,1024);
    sprintf (name, "xAMp Monitor @ %s", hostname);
    argv[0]=name;
    
    toplevel = XtAppInitialize (&app_con, "XXamp",
				options, XtNumber (options),
				&argc, argv, fallback_resources,
				NULL, ZERO);
    
    XtAppAddInput (app_con, xamp_fd, (XtPointer)XtInputReadMask, 
		   (XtInputCallbackProc)xampReadInput, NULL);

    depth = XDisplayPlanes(XtDisplay(toplevel),
                               DefaultScreen(XtDisplay(toplevel)));

    gray1 =
	XCreatePixmapFromBitmapData (XtDisplay(toplevel),
				     RootWindowOfScreen(XtScreen(toplevel)),
				     (char *)gray1_bits, 
				     gray1_width, gray1_height,
				     BlackPixel(XtDisplay(toplevel),
					 DefaultScreen(XtDisplay(toplevel))),
				     WhitePixel(XtDisplay(toplevel),
					 DefaultScreen(XtDisplay(toplevel))),
				     depth);
    
    box = XtVaCreateManagedWidget ("xxamp", formWidgetClass, toplevel,
				   XtNbackgroundPixmap, gray1,
				   NULL);
    
    
    emptypix =
	XCreatePixmapFromBitmapData (XtDisplay(toplevel),
				     RootWindowOfScreen(XtScreen(toplevel)),
				     (char *)mgsstate_empty_bits,
				     mgsstate_empty_width, 
				     mgsstate_empty_height,
				     BlackPixel(XtDisplay(toplevel),
					 DefaultScreen(XtDisplay(toplevel))),
				     WhitePixel(XtDisplay(toplevel),
					 DefaultScreen(XtDisplay(toplevel))),
				     depth);
    
    activepix =
	XCreatePixmapFromBitmapData (XtDisplay(toplevel),
				     RootWindowOfScreen(XtScreen(toplevel)),
				     (char *)mgsstate_active_bits,
				     mgsstate_active_width, 
				     mgsstate_active_height,
				     BlackPixel(XtDisplay(toplevel),
					 DefaultScreen(XtDisplay(toplevel))),
				     WhitePixel(XtDisplay(toplevel),
					 DefaultScreen(XtDisplay(toplevel))),
				     depth);
    
    inactivepix =
	XCreatePixmapFromBitmapData (XtDisplay(toplevel),
				     RootWindowOfScreen(XtScreen(toplevel)),
				     (char *)mgsstate_inactive_bits,
				     mgsstate_inactive_width, 
				     mgsstate_inactive_height,
				     BlackPixel(XtDisplay(toplevel),
					 DefaultScreen(XtDisplay(toplevel))),
				     WhitePixel(XtDisplay(toplevel),
					 DefaultScreen(XtDisplay(toplevel))),
				     depth);

    /******
      General button widgets
    ******/
     
    quitbutton = XtVaCreateManagedWidget ("mbutton", commandWidgetClass, box, 
					  XtNlabel, "Quit",
					  NULL);
   
    mgsmenubutton = XtVaCreateManagedWidget ("mbutton", menuButtonWidgetClass, 
					     box,
					     XtNlabel, "MGS",
					     XtNfromHoriz, quitbutton,
					     NULL);

    mgsmenu = XtVaCreatePopupShell ("menu", simpleMenuWidgetClass, 
				    mgsmenubutton,
				    NULL);


    mgsjoinbutton = XtVaCreateManagedWidget ("entry", smeBSBObjectClass,
					     mgsmenu,
					     XtNlabel, " Join",
					     NULL);
    
    mgsleavebutton = XtVaCreateManagedWidget ("entry", smeBSBObjectClass,
					      mgsmenu,
					      XtNlabel, " Leave",
					      NULL);

    mgscheckbutton = XtVaCreateManagedWidget ("entry", smeBSBObjectClass,
					      mgsmenu,
					      XtNlabel, " Check", 
					      NULL);
    
    /*********
      Time service widgets
    **********/

    tsgetbutton = XtVaCreateManagedWidget ("mbutton", commandWidgetClass, 
					    box,
					    XtNlabel, "TS get",
					    XtNfromHoriz, mgsmenubutton,
					    NULL);

    tstimelabel = XtVaCreateManagedWidget ("output", labelWidgetClass, box,
					   XtNlabel, "0.0",
					   XtNfromHoriz, tsgetbutton,
					   XtNwidth, 93,
					   XtNjustify, XtJustifyRight,
					   NULL);
    


    timemenubutton = XtVaCreateManagedWidget("mbutton", menuButtonWidgetClass, 
					     box,
					     XtNlabel, "Timing",
					     XtNwidth, 55,
					     XtNfromHoriz, tstimelabel,
					     NULL);

    timemenu = XtVaCreatePopupShell ("menu", simpleMenuWidgetClass, 
				     timemenubutton,
				     NULL);


    timenonebutton = XtVaCreateManagedWidget ("entry", smeBSBObjectClass,
					      timemenu,
					      XtNsensitive, False,
					      XtNlabel, " None",
					      NULL);
    
    timeappbutton = XtVaCreateManagedWidget ("entry", smeBSBObjectClass,
					     timemenu,
					     XtNlabel, " App",
					     NULL);

    timedaemonbutton = XtVaCreateManagedWidget ("entry", smeBSBObjectClass,
						timemenu,
						XtNlabel, " Daemon", 
						NULL);






    timeform = XtVaCreateManagedWidget ("timeform", formWidgetClass, box,
					XtNdefaultDistance, 0,
					XtNfromHoriz, timemenubutton,
					NULL);
    
    conflabel = XtVaCreateManagedWidget ("timelabel", labelWidgetClass, 
					 timeform,
					 XtNlabel, "Conf:",
					 XtNsensitive, False,
					 XtNjustify, XtJustifyRight,
					 XtNborderWidth, 0,
					NULL);

    confresultbutton = XtVaCreateManagedWidget ("output", labelWidgetClass, 
						timeform,
						XtNlabel, "0.0",
						XtNsensitive, False,
						XtNborderWidth, 0,
						XtNjustify, XtJustifyLeft,
						XtNfromHoriz, conflabel,
						NULL);

    indlabel = XtVaCreateManagedWidget ("timelabel", labelWidgetClass, 
					timeform,
					XtNfromHoriz, confresultbutton,
					XtNlabel, "Indi:",
					XtNsensitive, False,
					XtNjustify, XtJustifyRight,
					XtNborderWidth, 0,
					NULL);

    indresultbutton = XtVaCreateManagedWidget ("output", labelWidgetClass, 
					       timeform,
					       XtNlabel, "0.0",
					       XtNsensitive, False,
					       XtNborderWidth, 0,
					       XtNjustify, XtJustifyLeft,
					       XtNfromHoriz, indlabel,
					       NULL);

    /******
      Mgs table widgets
    ******/
    
    mgsstatepaned = XtVaCreateManagedWidget ("paned", panedWidgetClass, box,
					     XtNwidth, 125,
					     XtNinternalBorderWidth, 0,
					     XtNfromVert, quitbutton,
					     NULL);
    
    XtVaCreateManagedWidget("box", labelWidgetClass, mgsstatepaned, 
			    XtNborderWidth, 1,
			    XtNshowGrip, False,
			    XtNlabel, "MGS State",
			    NULL);

    for (i=0; i < MAX_STATIONS; i++) 
    {
	mgsstate[i] = XtVaCreateManagedWidget ("mgslabel", labelWidgetClass,  
					       mgsstatepaned, 
					       XtNborderWidth, 0,
					       XtNshowGrip, False,
					       XtNlabel, "-",
					       XtNleftBitmap, emptypix,
					       XtNjustify, XtJustifyLeft,
					       NULL);
    }
    
    /******
      Info window widget
    ******/

    mgstext = XtVaCreateManagedWidget ("infotext", asciiTextWidgetClass, box,
				       XtNfromVert, mgsstatepaned,
				       XtNscrollVertical, XawtextScrollAlways,
				       XtNeditType, XawtextAppend,
				       XtNstring, "",
				       NULL);

    
    /******
      Data input widgets
    ******/
    
    datactl = XtVaCreateManagedWidget ("form", formWidgetClass, box,
					XtNfromHoriz, mgsstatepaned,
					XtNfromVert, quitbutton,
					NULL);
    
    grouplabel = XtVaCreateManagedWidget ("datalabel", labelWidgetClass, 
					  datactl,
					  XtNlabel, "GrpId:",
					  NULL);

    groupidbutton = XtVaCreateManagedWidget ("datatext", asciiTextWidgetClass,
					     datactl, 
					     XtNfromHoriz, grouplabel,
					     XtNeditType, XawtextAppend,
					     XtNwidth, 60,
					     XtNstring, "0",
					     XtNinsertPosition, 1,
					     NULL);

    objlabel = XtVaCreateManagedWidget ("datalabel", labelWidgetClass, datactl,
					XtNfromVert, grouplabel,
					XtNlabel, "ObjId:",
					NULL);
    
    objidbutton = XtVaCreateManagedWidget ("datatext", asciiTextWidgetClass,
					   datactl, 
					   XtNfromHoriz, objlabel,
					   XtNfromVert, groupidbutton,
					   XtNeditType, XawtextAppend,
					   XtNwidth, 60,
					   XtNstring, "0",
					   XtNinsertPosition, 1,
					   NULL);

    sizelabel = XtVaCreateManagedWidget ("datalabel", labelWidgetClass, 
					 datactl, 
					 XtNfromHoriz, objidbutton,
					 XtNlabel, "Size:",
					 NULL);
    
    sizebutton = XtVaCreateManagedWidget ("datatext", asciiTextWidgetClass,
					  datactl, 
					  XtNfromHoriz, sizelabel,
					  XtNeditType, XawtextAppend,
					  NULL);

    priolabel = XtVaCreateManagedWidget ("datalabel", labelWidgetClass, 
					 datactl,
					 XtNfromHoriz, objidbutton,
					 XtNfromVert, sizelabel,
					 XtNlabel, "Prio:",
					 NULL);
    
    priobutton = XtVaCreateManagedWidget ("datatext", asciiTextWidgetClass,
					  datactl, 
					  XtNfromHoriz, priolabel,
					  XtNfromVert, sizebutton,
					  XtNeditType, XawtextAppend,
					  NULL);

    
    /******
      Command widgets
    ******/

    comctl = XtVaCreateManagedWidget ("form", formWidgetClass, box,
				      XtNfromHoriz, datactl,
				      XtNfromVert, quitbutton,
				      NULL);
   
    groupopenbutton = XtVaCreateManagedWidget ("button", commandWidgetClass,
					       comctl, 
					       XtNlabel, "Open",
					       NULL);
    
    groupclosebutton = XtVaCreateManagedWidget("button", commandWidgetClass,
					       comctl, 
					       XtNlabel, "Close",
					       XtNfromVert, groupopenbutton,
					       NULL);

    reliable = XtVaCreateManagedWidget ("button", commandWidgetClass, comctl, 
					XtNlabel, "Reliable",
					XtNfromHoriz, groupopenbutton,
					NULL);
    
    atomic = XtVaCreateManagedWidget ("button", commandWidgetClass, comctl,
				      XtNlabel,"Atomic",
				      XtNfromVert, reliable,
				      XtNfromHoriz, groupclosebutton,
				      NULL);
    
    tight = XtVaCreateManagedWidget ("button", commandWidgetClass, comctl, 
				     XtNlabel, "Tight",
				     XtNfromHoriz, reliable,
				     NULL);
    
    delta = XtVaCreateManagedWidget ("button", commandWidgetClass, comctl, 
				     XtNlabel, "Delta",
				     XtNfromVert, tight,
				     XtNfromHoriz, atomic,
				     NULL);
    
    
    /******
      Group membership table widgets
    ******/

    groupview = XtVaCreateManagedWidget ("formgv", formWidgetClass, box,
					 XtNdefaultDistance, 0,
					 XtNfromHoriz, mgsstatepaned,
					 XtNfromVert, comctl,
					 NULL);
    
    gvpanedlabel = XtVaCreateManagedWidget ("gvbox", panedWidgetClass, 
					    groupview,
					    XtNinternalBorderWidth, 0,
					    XtNborderWidth, 0,
					    NULL);
    gvpanedview = XtVaCreateManagedWidget("gvbox", panedWidgetClass, 
					  groupview,
					  XtNinternalBorderWidth, 0,
					  XtNborderWidth, 0,
					  XtNfromHoriz, gvpanedlabel,
					  NULL);
    
    for (i=0; i < 7; i++) 
    {
	groupstateview[i] = XtVaCreateManagedWidget("groupview", 
						    labelWidgetClass, 
						    gvpanedview, 
						    XtNborderWidth, 0,
						    XtNshowGrip, False,
						    XtNlabel, "",
						    XtNjustify, XtJustifyLeft,
						    XtNwidth, 351,
						    NULL);
	groupstatelabel[i] = XtVaCreateManagedWidget("grouplabel", 
						     labelWidgetClass, 
						     gvpanedlabel,
						     XtNborderWidth, 0,
						     XtNshowGrip, False,
						     XtNlabel, "",
						     XtNjustify, XtJustifyLeft,
						     XtNwidth, 50,
						     NULL);
    }
    
    XtSetArg (args[0], XtNlabel, " GroupId");
    XtSetArg (args[1], XtNjustify, XtJustifyLeft);
    XtSetArg (args[2], XtNheight, 40);
    XtSetValues (groupstatelabel[0], args, THREE);
    XtSetArg (args[0], XtNlabel, "View");
    XtSetValues (groupstateview[0], args, THREE);

   
    XtAddCallback (quitbutton,       XtNcallback, quitCallb,         NULL);
    XtAddCallback (mgsjoinbutton,    XtNcallback, mgsJoinCallb,      NULL);
    XtAddCallback (mgsleavebutton,   XtNcallback, mgsLeaveCallb,     NULL);
    XtAddCallback (mgscheckbutton,   XtNcallback, mgsCheckCallb,     NULL);
    XtAddCallback (timenonebutton,   XtNcallback, timingNoneCallb,   NULL);
    XtAddCallback (timeappbutton,    XtNcallback, timingAppCallb,    NULL);
    XtAddCallback (timedaemonbutton, XtNcallback, timingDaemonCallb, NULL);
    XtAddCallback (groupopenbutton,  XtNcallback, groupOpenCallb,    NULL);
    XtAddCallback (groupclosebutton, XtNcallback, groupCloseCallb,   NULL);
    XtAddCallback (reliable,         XtNcallback, reliableCallb,     NULL);
    XtAddCallback (atomic,           XtNcallback, atomicCallb,       NULL);
    XtAddCallback (tight,            XtNcallback, tightCallb,        NULL);
    XtAddCallback (delta,            XtNcallback, deltaCallb,        NULL);
    XtAddCallback (tsgetbutton,      XtNcallback, tsReadCallb,       NULL);

    XtRealizeWidget (toplevel);
    XtAppMainLoop (app_con);
}


